/* Creates a TCP server (client_tcp.c) 
 * 
 * Name: Ryan Prins
 * Date: Feb 19, 2004
 * INFO 341 - Lab 4
 *
 * Code portions borrowed from: http://netbook.cs.purdue.edu/code/code30_1.htm
 */

#ifndef unix
#define WIN32
#include <windows.h>
#include <winsock.h>
#else
#define closesocket close
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define PROTOPORT       5193            /* default protocol port number */

#define ERROR   0

#define END_LINE 0x0
#define MAX_MSG 100

extern  int             errno;
char    localhost[] =   "localhost";    /* default host name            */
/*------------------------------------------------------------------------
 * Program:   client
 *
 * Purpose:   allocate a socket, connect to a server, and print all output
 *
 * Syntax:    client [ host [port] ]
 *
 *               host  - name of a computer on which server is executing
 *               port  - protocol port number server is using
 *
 * Note:      Both arguments are optional.  If no host name is specified,
 *            the client uses "localhost"; if no protocol port is
 *            specified, the client uses the default given by PROTOPORT.
 *
 *------------------------------------------------------------------------
 */

int read_line(int, char *);

main(int argc, char *argv[])
{
    struct  hostent  *ptrh;  /* pointer to a host table entry           */
    struct  protoent *ptrp;  /* pointer to a protocol table entry       */
    struct  sockaddr_in sad; /* structure to hold an IP address         */
    int     sd;              /* socket descriptor                       */
    int     port;            /* protocol port number                    */
    char    *host;           /* pointer to host name                    */
    int     n;               /* number of characters read               */
    char    buf[1000];       /* buffer for data from the server         */
    char    command[1000];   /* buffer for command from user            */
    int     rc;              /* return code from send                   */
#ifdef WIN32
    WSADATA wsaData;
    WSAStartup(0x0101, &wsaData);
#endif
    memset((char *)&sad,0,sizeof(sad)); /* clear sockaddr structure     */
    sad.sin_family = AF_INET;           /* set family to Internet       */

    /* Check command-line argument for protocol port and extract        */
    /* port number if one is specified.  Otherwise, use the default     */
    /* port value given by constant PROTOPORT                           */

    if (argc > 2) {                     /* if protocol port specified   */
        port = atoi(argv[2]);           /* convert to binary            */
    } else {
        port = PROTOPORT;               /* use default port number      */
    }
    if (port > 0)                       /* test for legal value         */
        sad.sin_port = htons((u_short)port);
    else {                              /* print error message and exit */
        fprintf(stderr,"bad port number %s\n",argv[2]);
        exit(1);
    }

    /* Check host argument and assign host name.                        */
    if (argc > 1) {
        host = argv[1];                 /* if host argument specified   */
    } else {
        host = localhost;
    }

    /* Convert host name to equivalent IP address and copy to sad.      */
    ptrh = gethostbyname(host);
    if ( ((char *)ptrh) == NULL ) {
        fprintf(stderr,"invalid host: %s\n", host);
        exit(1);
    }
    memcpy(&sad.sin_addr, ptrh->h_addr, ptrh->h_length);

    /* Map TCP transport protocol name to protocol number.              */
    if ( ((int)(ptrp = getprotobyname("tcp"))) == 0) {
        fprintf(stderr, "cannot map \"tcp\" to protocol number");
        exit(1);
    }

    /* Create a socket.                                                 */
    sd = socket(PF_INET, SOCK_STREAM, ptrp->p_proto);
    if (sd < 0) {
        fprintf(stderr, "socket creation failed\n");
        exit(1);
    }

    /* Connect the socket to the specified server.                      */
    if (connect(sd, (struct sockaddr *)&sad, sizeof(sad)) < 0) {
        fprintf(stderr,"connect failed\n");
        exit(1);
    }
        
    printf("Client: ");
    gets(command);
    while(strcmp(command, ".quit") != 0) {

        /* Send a line of text to the server                            */
        rc = send(sd, command, strlen(command) + 1, 0);
	if(rc < 0) {
	    perror("Cannot send message\n");
	    close(sd);
	    exit(1);
	}

        /* Repeatedly read data from socket and write to user's screen. */
	if(read_line(sd, buf) <= 0) {
	    perror("Server disconnected\n");
	    closesocket(sd);
            exit(1);
        }

	printf("Server: %s\n", buf);

        printf("Client: ");
        gets(command);
    }

    /* Close the socket. */
    closesocket(sd);
}

int read_line(int newSd, char *line_to_return) {
  
    static int read_line_rcv_ptr=0;
    static char read_line_rcv_msg[MAX_MSG];
    static int read_line_n;
    int offset;

    offset=0;

    while(1) {
        if(read_line_rcv_ptr==0) {
            /* read data from socket */
            memset(read_line_rcv_msg,0x0,MAX_MSG); /* init buffer */
            read_line_n = recv(newSd, read_line_rcv_msg, MAX_MSG, 0); /* wait for data */
            if (read_line_n < 0) {
                perror("Cannot receive data\n");
                return ERROR;
            } else if (read_line_n == 0) {
                printf("Connection closed by server\n");
                close(newSd);
                return ERROR;
            }
        }
  
        /* if new data read on socket */
        /* OR */
        /* if another line is still in buffer */

        /* copy line into 'line_to_return' */
        while(*(read_line_rcv_msg+read_line_rcv_ptr) != END_LINE && read_line_rcv_ptr<read_line_n) {
            memcpy(line_to_return+offset,read_line_rcv_msg+read_line_rcv_ptr,1);
            offset++;
            read_line_rcv_ptr++;
        }
    
        /* end of line + end of buffer => return line */
        if(read_line_rcv_ptr == read_line_n-1) { 
            /* set last byte to END_LINE */
            *(line_to_return+offset)=END_LINE;
            read_line_rcv_ptr=0;
            return ++offset;
        } 
    
        /* end of line but still some data in buffer => return line */
        if(read_line_rcv_ptr < read_line_n-1) {
            /* set last byte to END_LINE */
            *(line_to_return+offset)=END_LINE;
            read_line_rcv_ptr++;
            return ++offset;
        }

        /* end of buffer but line is not ended => */
        /*  wait for more data to arrive on socket */
        if(read_line_rcv_ptr == read_line_n) {
            read_line_rcv_ptr = 0;
        } 
    
    } /* while */
}
