/* Creates a TCP server (server_tcp.c) 
 * 
 * Name: Ryan Prins
 * Date: Feb 19, 2004
 * INFO 341 - Lab 4
 *
 * Code portions borrowed from: http://netbook.cs.purdue.edu/code/code30_2.htm
 */

#ifndef unix
#define WIN32
#include <windows.h>
#include <winsock.h> 
#else
#define closesocket close
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#endif

#include <stdio.h>
#include <string.h>

#define PROTOPORT       5193            /* default protocol port number */
#define QLEN            6               /* size of request queue        */

#define ERROR   0

#define END_LINE 0x0
#define MAX_MSG 100

int     visits      =   0;              /* counts client connections    */
/*------------------------------------------------------------------------
 * Program:   server
 *
 * Purpose:   allocate a socket and then repeatedly execute the following:
 *              (1) wait for the next connection from a client
 *              (2) send a short message to the client
 *              (3) close the connection
 *              (4) go back to step (1)
 *
 * Syntax:    server [ port ]
 *
 *               port  - protocol port number to use
 *
 * Note:      The port argument is optional.  If no port is specified,
 *            the server uses the default given by PROTOPORT.
 *
 *			  Server will quit when .quit is typed into console
 *
 *------------------------------------------------------------------------
 */

int read_line(int, char *);

main(int argc, char *argv[])
{
    struct  hostent  *ptrh;  /* pointer to a host table entry       */
    struct  protoent *ptrp;  /* pointer to a protocol table entry   */
    struct  sockaddr_in sad; /* structure to hold server's address  */
    struct  sockaddr_in cad; /* structure to hold client's address  */
    int     sd, sd2;         /* socket descriptors                  */
    int     port;            /* protocol port number                */
    int     alen;            /* length of address                   */
    char    buf[1000];       /* buffer for string the server sends  */
    char    command[1000];   /* buffer for command from user        */
    int     rc;              /* return code from send               */
    int     clientconnected; /* whether a client is connected       */
#ifdef WIN32
    WSADATA wsaData;
    WSAStartup(0x0101, &wsaData);
#endif
    memset((char *)&sad,0,sizeof(sad)); /* clear sockaddr structure */
    sad.sin_family = AF_INET;         /* set family to Internet     */
    sad.sin_addr.s_addr = INADDR_ANY; /* set the local IP address   */

    /* Check command-line argument for protocol port and extract    */
    /* port number if one is specified.  Otherwise, use the default */
    /* port value given by constant PROTOPORT                       */

    if (argc > 1) {                 /* if argument specified        */
        port = atoi(argv[1]);   /* convert argument to binary   */
    } else {
        port = PROTOPORT;       /* use default port number      */
    }
    if (port > 0)                   /* test for illegal value       */
        sad.sin_port = htons((u_short)port);
    else {                          /* print error message and exit */
        fprintf(stderr,"bad port number %s\n",argv[1]);
        exit(1);
    }

    /* Map TCP transport protocol name to protocol number */
    if ( ((int)(ptrp = getprotobyname("tcp"))) == 0) {
        fprintf(stderr, "cannot map \"tcp\" to protocol number");
        exit(1);
    }

    /* Create a socket */
    sd = socket(PF_INET, SOCK_STREAM, ptrp->p_proto);
    if (sd < 0) {
        fprintf(stderr, "socket creation failed\n");
        exit(1);
    }

    /* Bind a local address to the socket */
    if (bind(sd, (struct sockaddr *)&sad, sizeof(sad)) < 0) {
        fprintf(stderr,"bind failed\n");
        exit(1);
    }

    /* Specify size of request queue */
    if (listen(sd, QLEN) < 0) {
        fprintf(stderr,"listen failed\n");
        exit(1);
    }

    /* Main server loop - accept and handle requests */
    while (1) {
        alen = sizeof(cad);
        if ( (sd2=accept(sd, (struct sockaddr *)&cad, &alen)) < 0) {
            fprintf(stderr, "accept failed\n");
            exit(1);
        }
        visits++;
	clientconnected = 1;

	/* Repeatedly read data from socket and write to user's screen. */
	if(read_line(sd2, buf) <= 0) {
	    perror("Client disconnected\n");
	    clientconnected = 0;
	}
	else {
	    printf("Client: %s\n", buf);
	}

        while(clientconnected == 1) {
            printf("Server: ");
            gets(command);
			/* Close the socket. If user enters in .quit */
            if(strcmp(command, ".quit") == 0) {
		
				closesocket(sd);
				exit(0);
				}
	    
            /* Send a line of text to the server                            */
            rc = send(sd2, command, strlen(command) + 1, 0);
            if(rc < 0) {
                perror("Cannot send message\n");
                clientconnected = 0;
            }

            /* Repeatedly read data from socket and write to user's screen. */
            if(read_line(sd2, buf) <= 0) {
                perror("Client disconnected\n");
		clientconnected = 0;
            }
	    else {
		printf("Client: %s\n", buf);
	    }
        }
    }
}

int read_line(int newSd, char *line_to_return) {
  
    static int read_line_rcv_ptr=0;
    static char read_line_rcv_msg[MAX_MSG];
    static int read_line_n;
    int offset;

    offset=0;

    while(1) {
        if(read_line_rcv_ptr==0) {
            /* read data from socket */
            memset(read_line_rcv_msg,0x0,MAX_MSG); /* init buffer */
            read_line_n = recv(newSd, read_line_rcv_msg, MAX_MSG, 0); /* wait for data */
            if (read_line_n < 0) {
                perror("Cannot receive data\n");
                return ERROR;
            } else if (read_line_n == 0) {
                printf("Connection closed by client\n");
                close(newSd);
                return ERROR;
            }
        }
  
        /* if new data read on socket */
        /* OR */
        /* if another line is still in buffer */

        /* copy line into 'line_to_return' */
        while(*(read_line_rcv_msg+read_line_rcv_ptr) != END_LINE && read_line_rcv_ptr<read_line_n) {
            memcpy(line_to_return+offset,read_line_rcv_msg+read_line_rcv_ptr,1);
            offset++;
            read_line_rcv_ptr++;
        }
    
        /* end of line + end of buffer => return line */
        if(read_line_rcv_ptr == read_line_n-1) { 
            /* set last byte to END_LINE */
            *(line_to_return+offset)=END_LINE;
            read_line_rcv_ptr=0;
            return ++offset;
        } 
    
        /* end of line but still some data in buffer => return line */
        if(read_line_rcv_ptr < read_line_n-1) {
            /* set last byte to END_LINE */
            *(line_to_return+offset)=END_LINE;
            read_line_rcv_ptr++;
            return ++offset;
        }

        /* end of buffer but line is not ended => */
        /*  wait for more data to arrive on socket */
        if(read_line_rcv_ptr == read_line_n) {
            read_line_rcv_ptr = 0;
        } 
    
    } /* while */
}
